/*  Copyright (C) 2008--2010 Joshua Judson Rosen <rozzin@geekspace.com>.

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    version 3 as published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; see the file COPYING. If not, see
    <http://www.gnu.org/licenses/> or write to:

        The Free Software Foundation, inc.
        51 Franklin Street, Fifth Floor
        Boston, MA 02110-1301
        USA
*/

#include <stdlib.h>
#include "constants.h"
#include "random.h"

#include <math.h>
#include <cairo.h>

#include <glib-object.h>
#include "graphics.h"
#include "line.h"

#include <glib.h>

struct _VisualID_LineClass {
  VisualID_GlyphClass parent_class;
};

G_DEFINE_TYPE(VisualID_Line, visualid_line, VISUALID_TYPE_GLYPH)

static void
exec_line (VisualID_Line *line, cairo_t *cr)
{
  double x, y;
  int i;

  cairo_get_current_point (cr, &x, &y);
  cairo_new_sub_path (cr);

  cairo_move_to (cr, 0, line->len/2);
  if (line->ec1) {
    cairo_save (cr);
    cairo_translate (cr, 0, line->len/2);
    cairo_scale (cr, line->c1_scale, line->c1_scale);
    visualid_draw_path (line->ec1, cr);
    cairo_restore (cr);
  }

  cairo_line_to (cr, 0, -line->len/2);
  if (line->ec2) {
    cairo_save (cr);
    cairo_translate (cr, 0, -line->len/2);
    cairo_scale (cr, line->c2_scale, line->c2_scale);
    visualid_draw_path (line->ec2, cr);
    cairo_restore (cr);
  }

  if (line->c3) {
    for (i = 0;
         i < line->n;
         i++) {
      /* # of 0-based indices along the line;
         or the number of *spaces* between ribs: */
      int nspaces = (line->n - 1);

      /* Shift the scale so that, rather than covering [0%, 100%],
         it covers [-50%, 50%] with zero at the centre: */
      double base_offset = -.5;

      double progress = 0;
      if (nspaces != 0) {
        progress = (double) i / nspaces;
      }

      cairo_save (cr);
      cairo_translate (cr, 0, line->len * (base_offset + progress));
      cairo_scale (cr, line->c3_scale, line->c3_scale);
      if (line->c3_taper) {
        double taper = 1 + (.5 * i/line->n);
        if (!line->c3_increase) {
          taper = 1/taper;
        }
        cairo_scale (cr, taper, taper);
      }
      if (line->c3_doublesided) {
        cairo_translate (cr, 1, 0);
        cairo_save (cr);
        cairo_rotate (cr, line->c3_angle - M_PI_2);
        visualid_draw_path (line->c3, cr);
        cairo_restore (cr);
        cairo_scale (cr, -1, 1);
        cairo_translate (cr, 2, 0);
      }
      cairo_rotate (cr, line->c3_angle - M_PI_2);
      visualid_draw_path (line->c3, cr);
      cairo_restore (cr);
    }
  }

  cairo_new_sub_path (cr);
  cairo_move_to (cr, x, y);
}

VisualID_Glyph *
gen_line (VisualID_Glyph *parent, const char *attachment)
{
  int level = parent ? parent->recursion_level+1 : 0;

  return  g_object_new (VISUALID_TYPE_LINE,
                       "recursion-level", level,
                       "parent-glyph", parent,
                       "parent-attachment", attachment,
                       NULL);
}

static void
line_constructed (GObject *self)
{
  VisualID_Line *line = VISUALID_LINE (self);
  VisualID_Glyph *glyph = VISUALID_GLYPH (self);
  GObjectClass *parent_class = G_OBJECT_CLASS (visualid_line_parent_class);

  int level = glyph->recursion_level;

  if (parent_class->constructed) {
    parent_class->constructed (self);
  }

  line->len = rndin (0.5, 1);
  line->n = rndin_lowbias (3, 10, 1.5);

  if (level < recursion_max && rndprob (0.5)) {
    visualid_glyph_spawn (glyph, "rib-child");
    line->c3_scale = rndin_lowbias (0.05, 0.7, 1.5);
    line->c3_doublesided = rndprob (0.8);
    line->c3_angle = rndin (-M_PI_4, M_PI_4);
    line->c3_taper = rndprob (0.3);
    line->c3_increase = rndprob (0.3);
  } else {
    line->c3 = NULL;
  }

  if (level < recursion_max && rndprob (0.75)) {
    visualid_glyph_spawn (glyph, "tail-child");
    line->c2_scale = rndin (0.15, 0.6);
  } else {
    line->ec2 = NULL;
  }

  if (level < recursion_max && (!(line->ec2 || line->c3) || rndprob (0.75)))
  {
    visualid_glyph_spawn (glyph, "head-child");
    line->c1_scale = rndin (0.15, 0.6);
  } else {
    line->ec1 = NULL;
  }
}

static void
line_dispose (GObject *self)
{
  VisualID_Line *line = VISUALID_LINE (self);

  if (line->ec1) {
    visualid_glyph_unparent (line->ec1);
    line->ec1 = NULL;
  }

  if (line->ec2) {
    visualid_glyph_unparent (line->ec2);
    line->ec2 = NULL;
  }

  if (line->c3) {
    visualid_glyph_unparent (line->c3);
    line->c3 = NULL;
  }

  G_OBJECT_CLASS (visualid_line_parent_class)->dispose (self);
}

enum {
  PROP_LENGTH = 1,

  PROP_HEAD_CHILD,
  PROP_HEAD_SCALE,

  PROP_TAIL_CHILD,
  PROP_TAIL_SCALE,

  PROP_RIB_CHILD,
  PROP_RIB_COUNT,
  PROP_RIB_SCALE,
  PROP_RIB_ANGLE,
  PROP_SYMMETRY,
  PROP_TAPER,
  PROP_INCREASE
};

static void
line_get_property (GObject *object,
                   guint property_id,
                   GValue *value,
                   GParamSpec *pspec)
{
  VisualID_Line *self = VISUALID_LINE (object);

  switch (property_id)
    {
    case PROP_LENGTH:
      g_value_set_double (value, self->len);
      break;
    case PROP_HEAD_CHILD:
      g_value_set_object (value, self->ec1);
      break;
    case PROP_HEAD_SCALE:
      g_value_set_double (value, self->c1_scale);
      break;

    case PROP_TAIL_CHILD:
      g_value_set_object (value, self->ec2);
      break;
    case PROP_TAIL_SCALE:
      g_value_set_double (value, self->c2_scale);
      break;

    case PROP_RIB_COUNT:
      g_value_set_int (value, self->n);
      break;
    case PROP_RIB_SCALE:
      g_value_set_double (value, self->c3_scale);
      break;
    case PROP_RIB_ANGLE:
      g_value_set_double (value, self->c3_angle);
      break;
    case PROP_RIB_CHILD:
      g_value_set_object (value, self->c3);
      break;
    case PROP_SYMMETRY:
      g_value_set_boolean (value, self->c3_doublesided);
      break;
    case PROP_TAPER:
      g_value_set_boolean (value, self->c3_taper);
      break;

    case PROP_INCREASE:
      g_value_set_boolean (value, self->c3_increase);

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
line_set_property (GObject *object,
                   guint property_id,
                   const GValue *value,
                   GParamSpec *pspec)
{
  VisualID_Line *self = VISUALID_LINE (object);
  switch (property_id)
  {
    case PROP_LENGTH:
      self->len = g_value_get_double (value);
      break;

    case PROP_HEAD_CHILD:
      visualid_glyph_take_child (VISUALID_GLYPH (self),
                                 g_value_get_object (value),
                                 &self->ec1);
      break;
    case PROP_HEAD_SCALE:
      self->c1_scale = g_value_get_double (value);
      break;

    case PROP_TAIL_CHILD:
      visualid_glyph_take_child (VISUALID_GLYPH (self),
                                 g_value_get_object (value),
                                 &self->ec2);
      break;
    case PROP_TAIL_SCALE:
      self->c2_scale = g_value_get_double (value);
      break;

    case PROP_RIB_CHILD:
      visualid_glyph_take_child (VISUALID_GLYPH (self),
                                 g_value_get_object (value),
                                 &self->c3);
      break;
    case PROP_RIB_COUNT:
      self->n = g_value_get_int (value);
      break;
    case PROP_RIB_SCALE:
      self->c3_scale = g_value_get_double (value);
      break;
    case PROP_RIB_ANGLE:
      self->c3_angle = g_value_get_double (value);
      break;
    case PROP_SYMMETRY:
      self->c3_doublesided = g_value_get_boolean (value);
      break;
    case PROP_TAPER:
      self->c3_taper = g_value_get_boolean (value);
      break;
    case PROP_INCREASE:
      self->c3_increase = g_value_get_boolean (value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
  }
}

static int
visualid_line_complexity (VisualID_Line *line, int recurse)
{
  int complexity = 1;

  if (recurse--) {
    if (line->ec1) {
      complexity += visualid_glyph_complexity (line->ec1, recurse);
    }

    if (line->ec2) {
      complexity += visualid_glyph_complexity (line->ec2, recurse);
    }

    if (line->c3) {
      int rib_complexity = visualid_glyph_complexity (line->c3, recurse);

      if (line->c3_doublesided) {
        rib_complexity *= 2;
      }

      complexity += (rib_complexity * line->n);
    }
  }

  return complexity;
}

static void
visualid_line_class_init (VisualID_LineClass *class)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (class);

  VISUALID_GLYPH_CLASS (class)->produce = (producer_fn *) exec_line;
  VISUALID_GLYPH_CLASS (class)->complexity_fn =
    (complexity_fn *) visualid_line_complexity;

  gobject_class->get_property = line_get_property;
  gobject_class->set_property = line_set_property;
  gobject_class->constructed = line_constructed;
  gobject_class->dispose = line_dispose;

  g_object_class_install_property
    (gobject_class, PROP_LENGTH,
     g_param_spec_double ("line-length",
                          "Line length",
                          "Relative length of the line",
                          .5, 1, .75,
                          G_PARAM_READWRITE));

  g_object_class_install_property
    (gobject_class, PROP_HEAD_CHILD,
     g_param_spec_object ("head-child",
                          "Head child",
                          "Glyph to place at the top (or outward-pointing) end of the line",
                          VISUALID_TYPE_GLYPH,
                          G_PARAM_READWRITE));

  g_object_class_install_property
    (gobject_class, PROP_HEAD_SCALE,
     g_param_spec_double ("head-scale",
                          "Head scale",
                          "Relative scale of the head-glyph",
                          .15, .6, .4,
                          G_PARAM_READWRITE));

  g_object_class_install_property
    (gobject_class, PROP_TAIL_CHILD,
     g_param_spec_object ("tail-child",
                          "Tail child",
                          "Glyph to place at the bottom (or inward-pointing) end of the line",
                          VISUALID_TYPE_GLYPH,
                          G_PARAM_READWRITE));

  g_object_class_install_property
    (gobject_class, PROP_TAIL_SCALE,
     g_param_spec_double ("tail-scale",
                          "Tail scale",
                          "Relative scale of tail-glyph",
                          .15, .6, .4,
                          G_PARAM_READWRITE));

  g_object_class_install_property
    (gobject_class, PROP_RIB_CHILD,
     g_param_spec_object ("rib-child",
                          "Rib child",
                          "Glyph to place at regular intervals along the line",
                          VISUALID_TYPE_GLYPH,
                          G_PARAM_READWRITE));

  g_object_class_install_property
    (gobject_class, PROP_RIB_COUNT,
     g_param_spec_int ("rib-count",
                       "Rib count",
                       "The number of `rib' glyphs to place along the line",
                       3, 10, 5,
                       G_PARAM_READWRITE));

  g_object_class_install_property
    (gobject_class, PROP_RIB_SCALE,
     g_param_spec_double ("rib-scale",
                          "Rib scale",
                          "(Nominal) relative scale of rib-glyphs",
                          .05, .7, .1,
                          G_PARAM_READWRITE));

  g_object_class_install_property
    (gobject_class, PROP_RIB_ANGLE,
     g_param_spec_double ("rib-angle",
                          "Rib angle",
                          "Angle at which rib-glyphs are rotated relative to the line",
                          -M_PI_4, M_PI_4, 0,
                          G_PARAM_READWRITE));
}

static void
visualid_line_init (VisualID_Line *self)
{
}
